#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import io
import sys
import unittest.mock
import urllib.request

import pytest

import sambacc.rados_opener

# CAUTION: nearly all of these tests are based on mocking the ceph rados API.
# Testing this for real would require an operational ceph cluster and that's
# not happening for a simple set of unit tests!


def test_enable_rados_url_opener(monkeypatch):
    mock = unittest.mock.MagicMock()
    monkeypatch.setitem(sys.modules, "rados", mock)

    cls_mock = unittest.mock.MagicMock()
    sambacc.rados_opener.enable_rados(cls_mock)
    assert cls_mock._handlers.append.called


def test_enable_rados_url_opener_fail(monkeypatch):
    cls_mock = unittest.mock.MagicMock()
    sambacc.rados_opener.enable_rados(cls_mock)
    assert not cls_mock._handlers.append.called


def test_enable_rados_url_opener_with_args(monkeypatch):
    mock = unittest.mock.MagicMock()
    monkeypatch.setitem(sys.modules, "rados", mock)

    cls_mock = unittest.mock.MagicMock()
    cls_mock._handlers = []
    sambacc.rados_opener.enable_rados(cls_mock, client_name="user1")
    assert len(cls_mock._handlers) == 1
    assert isinstance(
        cls_mock._handlers[0]._interface, sambacc.rados_opener._RADOSInterface
    )
    assert cls_mock._handlers[0]._interface.api is mock
    assert cls_mock._handlers[0]._interface.client_name == "user1"
    assert not cls_mock._handlers[0]._interface.full_name
    ri = cls_mock._handlers[0]._interface
    ri.Rados()
    assert ri.api.Rados.call_args[1]["rados_id"] == "user1"
    assert ri.api.Rados.call_args[1]["name"] == ""
    assert (
        ri.api.Rados.call_args[1]["conffile"] == mock.Rados.DEFAULT_CONF_FILES
    )


def test_enable_rados_url_opener_with_args2(monkeypatch):
    mock = unittest.mock.MagicMock()
    monkeypatch.setitem(sys.modules, "rados", mock)

    cls_mock = unittest.mock.MagicMock()
    cls_mock._handlers = []
    sambacc.rados_opener.enable_rados(
        cls_mock, client_name="client.user1", full_name=True
    )
    assert len(cls_mock._handlers) == 1
    assert isinstance(
        cls_mock._handlers[0]._interface, sambacc.rados_opener._RADOSInterface
    )
    assert cls_mock._handlers[0]._interface.api is mock
    assert cls_mock._handlers[0]._interface.client_name == "client.user1"
    assert cls_mock._handlers[0]._interface.full_name
    ri = cls_mock._handlers[0]._interface
    ri.Rados()
    assert ri.api.Rados.call_args[1]["rados_id"] == ""
    assert ri.api.Rados.call_args[1]["name"] == "client.user1"
    assert (
        ri.api.Rados.call_args[1]["conffile"] == mock.Rados.DEFAULT_CONF_FILES
    )


def test_rados_handler_parse():
    class RH(sambacc.rados_opener._RADOSHandler):
        _rados_api = unittest.mock.MagicMock()

    rh = RH()
    rq = urllib.request.Request("rados://foo/bar/baz")
    rr = rh.rados_open(rq)
    assert rr._pool == "foo"
    assert rr._ns == "bar"
    assert rr._key == "baz"

    rq = urllib.request.Request("rados:///foo1/bar1/baz1")
    rr = rh.rados_open(rq)
    assert rr._pool == "foo1"
    assert rr._ns == "bar1"
    assert rr._key == "baz1"


def test_rados_handler_norados():
    # Generally, this shouldn't happen because the rados handler shouldn't
    # be added to the URLOpener if rados module was unavailable.
    class RH(sambacc.rados_opener._RADOSHandler):
        _interface = None

    rh = RH()
    rq = urllib.request.Request("rados://foo/bar/baz")
    with pytest.raises(sambacc.rados_opener.RADOSUnsupported):
        rh.rados_open(rq)


def test_rados_response_read_all():
    sval = b"Hello, World.\nI am a fake rados object.\n"

    def _read(_, size, off):
        if off < len(sval):
            return sval

    mock = unittest.mock.MagicMock()
    mock.Rados.return_value.open_ioctx.return_value.read.side_effect = _read

    rr = sambacc.rados_opener.RADOSObjectRef(mock, "foo", "bar", "baz")
    assert rr.readable()
    assert not rr.seekable()
    assert not rr.writable()
    assert not rr.isatty()
    assert rr.mode == "rb"
    assert rr.name == "baz"
    assert not rr.closed
    data = rr.read()
    assert data == sval
    assert rr.tell() == len(sval)
    rr.flush()
    rr.close()
    assert rr.closed


def test_rados_response_read_chunks():
    sval = b"a bad cat lives under the murky terrifying water"
    bio = io.BytesIO(sval)

    def _read(_, size, off):
        bio.seek(off)
        return bio.read(size)

    mock = unittest.mock.MagicMock()
    mock.Rados.return_value.open_ioctx.return_value.read.side_effect = _read

    rr = sambacc.rados_opener.RADOSObjectRef(mock, "foo", "bar", "baz")
    assert rr.readable()
    assert rr.read(8) == b"a bad ca"
    assert rr.read(8) == b"t lives "
    assert rr.read(8) == b"under th"


def test_rados_response_read_ctx_iter():
    sval = b"a bad cat lives under the murky terrifying water"
    bio = io.BytesIO(sval)

    def _read(_, size, off):
        bio.seek(off)
        return bio.read(size)

    mock = unittest.mock.MagicMock()
    mock.Rados.return_value.open_ioctx.return_value.read.side_effect = _read

    rr = sambacc.rados_opener.RADOSObjectRef(mock, "foo", "bar", "baz")
    with rr:
        result = [value for value in rr]
    assert result == [sval]
    with pytest.raises(ValueError):
        rr.read(8)


def test_rados_response_not_implemented():
    mock = unittest.mock.MagicMock()

    rr = sambacc.rados_opener.RADOSObjectRef(mock, "foo", "bar", "baz")
    with pytest.raises(NotImplementedError):
        rr.seek(10)
    with pytest.raises(NotImplementedError):
        rr.fileno()
    with pytest.raises(NotImplementedError):
        rr.readline()
    with pytest.raises(NotImplementedError):
        rr.readlines()
    with pytest.raises(NotImplementedError):
        rr.truncate()
    with pytest.raises(NotImplementedError):
        rr.write(b"zzzzz")
    with pytest.raises(NotImplementedError):
        rr.writelines([b"zzzzz"])


def test_rados_handler_config_key():
    class RH(sambacc.rados_opener._RADOSHandler):
        _interface = unittest.mock.MagicMock()

    mc = RH._interface.Rados.return_value.__enter__.return_value.mon_command
    mc.return_value = (0, b"rubber baby buggy bumpers", "")

    rh = RH()
    rq = urllib.request.Request("rados:mon-config-key:aa/bb/cc")
    rr = rh.rados_open(rq)
    assert isinstance(rr, io.BytesIO)
    assert rr.read() == b"rubber baby buggy bumpers"
    assert mc.called
    assert "aa/bb/cc" in mc.call_args[0][0]

    mc.reset_mock()
    mc.return_value = (2, b"", "no passing")
    rh = RH()
    rq = urllib.request.Request("rados:mon-config-key:xx/yy/zz")
    with pytest.raises(OSError) as pe:
        rh.rados_open(rq)
    assert getattr(pe.value, "errno", None) == 2
    assert mc.called
    assert "xx/yy/zz" in mc.call_args[0][0]



#
# sambacc: a samba container configuration tool
# Copyright (C) 2024  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from typing import Optional
import argparse
import os

from sambacc.typelets import Self

from .cli import Context


class SkipIf:
    """Base class for objects used to check if a particular sambacc command
    should be skipped.
    Skips are useful when different commands are chained together
    unconditionally in a configuration file (like k8s init containers) but
    certain commmands should not be run.
    """

    NAME: str = ""

    def test(self, ctx: Context) -> Optional[str]:
        """Return a string explaining the reason for the skip or None
        indicating no skip is desired.
        """
        raise NotImplementedError()  # pragma: nocover

    @classmethod
    def parse(cls, value: str) -> Self:
        """Parse a string into a skip class arguments."""
        raise NotImplementedError()  # pragma: nocover


class SkipFile(SkipIf):
    """Skip execution if a file exists or does not exist.
    The input value "file:/foo/bar" will trigger a skip if the file /foo/bar
    exists. To skip if a file does not exist, use "file:!/foo/bar" - prefix the
    file name with an exclaimation point.
    """

    NAME: str = "file"
    inverted: bool = False
    path: str = ""

    @classmethod
    def parse(cls, value: str) -> Self:
        obj = cls()
        if not value:
            raise ValueError("missing path")
        if value[0] == "!":
            obj.inverted = True
            value = value[1:]
        obj.path = value
        return obj

    def test(self, ctx: Context) -> Optional[str]:
        exists = os.path.exists(self.path)
        if self.inverted and not exists:
            return f"skip-if-file-missing: {self.path} missing"
        if not self.inverted and exists:
            return f"skip-if-file-exists: {self.path} exists"
        return None


class SkipEnv(SkipIf):
    """Skip execution if an environment variable is, or is not, equal to a
    value. The specification is roughly "env:<ENV_VAR><op><VALUE>" where op may
    be either `==` or `!=`. For example, "env:FLAVOR==cherry" will skip
    execution if the environment variable "FLAVOR" contains the value "cherry".
    "env:FLAVOR!=cherry" will skip execution if "FLAVOR" contains any value
    other than "cherry".
    """

    NAME: str = "env"
    _EQ = "=="
    _NEQ = "!="

    def __init__(self, op: str, var_name: str, value: str) -> None:
        self.op = op
        self.var_name = var_name
        self.target_value = value

    @classmethod
    def parse(cls, value: str) -> Self:
        if cls._EQ in value:
            op = cls._EQ
        elif cls._NEQ in value:
            op = cls._NEQ
        else:
            raise ValueError("invalid SkipEnv: missing or invalid operation")
        lhv, rhv = value.split(op, 1)
        return cls(op, lhv, rhv)

    def test(self, ctx: Context) -> Optional[str]:
        env_val = os.environ.get(self.var_name)
        if self.op == self._EQ and env_val == self.target_value:
            return (
                f"env var: {self.var_name}"
                f" -> {env_val} {self.op} {self.target_value}"
            )
        if self.op == self._NEQ and env_val != self.target_value:
            return (
                f"env var: {self.var_name}"
                f" -> {env_val} {self.op} {self.target_value}"
            )
        return None


class SkipAlways(SkipIf):
    """Skip execution unconditionally. Must be specified as "always:" and takes
    no value after the colon.
    """

    NAME: str = "always"

    @classmethod
    def parse(cls, value: str) -> Self:
        if value:
            raise ValueError("always skip takes no value")
        return cls()

    def test(self, ctx: Context) -> Optional[str]:
        return "always skip"


_SKIP_TYPES = [SkipFile, SkipEnv, SkipAlways]


def test(
    ctx: Context, *, conditions: Optional[list[SkipIf]] = None
) -> Optional[str]:
    """Return a string explaining the reason for a skip or None indicating
    no skip should be performed. Typically the skip conditions will be
    derived from the command line arguments but can be passed in manually
    using the `conditions` keyword argument.
    """
    if not conditions:
        conditions = ctx.cli.skip_conditions or []
    for cond in conditions:
        skip = cond.test(ctx)
        if skip:
            return skip
    return None


def parse(value: str) -> SkipIf:
    """Given a string return a SkipIf-based object. Every value must be
    prefixed with the skip "type" (the skip type's NAME).
    """
    if value == "?":
        # A hack to avoid putting tons of documentation into the help output.
        raise argparse.ArgumentTypeError(_help_info())
    for sk in _SKIP_TYPES:
        assert issubclass(sk, SkipIf)
        prefix = f"{sk.NAME}:"
        plen = len(prefix)
        if value.startswith(prefix):
            return sk.parse(value[plen:])
    raise KeyError("no matching skip rule for: {value!r}")


def _help_info() -> str:
    msgs = ["Skip conditions help details:", ""]
    for sk in _SKIP_TYPES:
        assert issubclass(sk, SkipIf)
        msgs.append(f"== Skip execution on condition `{sk.NAME}` ==")
        assert sk.__doc__
        for line in sk.__doc__.splitlines():
            msgs.append(line.strip())
        msgs.append("")
    return "\n".join(msgs)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import pytest

import sambacc.commands.cli
import sambacc.commands.main
from .test_netcmd_loader import config1


def run(*args):
    return sambacc.commands.main.main(args)


def test_no_id(capsys):
    with pytest.raises(sambacc.commands.cli.Fail):
        run("print-config")


def test_print_config(capsys, tmp_path):
    fname = tmp_path / "sample.json"
    with open(fname, "w") as fh:
        fh.write(config1)
    run("--identity", "foobar", "--config", str(fname), "print-config")
    out, err = capsys.readouterr()
    assert "[global]" in out
    assert "netbios name = GANDOLPH" in out
    assert "[share]" in out
    assert "path = /share" in out
    assert "[stuff]" in out
    assert "path = /mnt/stuff" in out


def test_print_config_env_vars(capsys, tmp_path, monkeypatch):
    fname = tmp_path / "sample.json"
    with open(fname, "w") as fh:
        fh.write(config1)
    monkeypatch.setenv("SAMBACC_CONFIG", str(fname))
    monkeypatch.setenv("SAMBA_CONTAINER_ID", "foobar")
    run("print-config")
    out, err = capsys.readouterr()
    assert "[global]" in out
    assert "netbios name = GANDOLPH" in out
    assert "[share]" in out
    assert "path = /share" in out
    assert "[stuff]" in out
    assert "path = /mnt/stuff" in out



#
# sambacc: a samba container configuration tool
# Copyright (C) 2022  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""typelets defines common-ish type hinting types that are tedious to
remember/redefine.
"""

from types import TracebackType
import sys
import typing

ExcType = typing.Optional[typing.Type[BaseException]]
ExcValue = typing.Optional[BaseException]
ExcTraceback = typing.Optional[TracebackType]


if sys.version_info >= (3, 11):
    from typing import Self
elif typing.TYPE_CHECKING:
    from typing_extensions import Self
else:
    Self = typing.Any



#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import typing


class SchemeNotSupported(Exception):
    pass


class Opener(typing.Protocol):
    """Protocol for a basic opener type that takes a path-ish or uri-ish
    string and tries to open it.
    """

    def open(self, path_or_uri: str) -> typing.IO:
        """Open a specified resource by path or (pseudo) URI."""
        ...  # pragma: no cover


class FallbackOpener:
    """FallbackOpener is used to open a path if a the string can not be
    opened as a URI/URL.
    """

    def __init__(
        self,
        openers: list[Opener],
        open_fn: typing.Optional[typing.Callable[..., typing.IO]] = None,
    ) -> None:
        self._openers = openers
        self._open_fn = open_fn or FileOpener.open

    def open(self, path_or_uri: str) -> typing.IO:
        for opener in self._openers:
            try:
                return opener.open(path_or_uri)
            except SchemeNotSupported:
                pass
        return self._open(path_or_uri)

    def _open(self, path: str) -> typing.IO:
        return self._open_fn(path)


class FileOpener:
    """Minimal opener that only supports opening local files."""

    @staticmethod
    def open(path: str) -> typing.IO:
        return open(path, "rb")



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import typing

from sambacc.typelets import ExcType, ExcValue, ExcTraceback


class LeaderStatus(typing.Protocol):
    """Fetches information about the current cluster leader."""

    def is_leader(self) -> bool:
        """Return true if the current node is the leader."""
        ...  # pragma: no cover


class LeaderLocator(typing.Protocol):
    """Acquire state needed to determine or fix a cluster leader.
    Can be used for purely informational types or types that
    actually acquire cluster leadership if needed.
    """

    def __enter__(self) -> LeaderStatus:
        """Enter context manager. Returns LeaderStatus."""
        ...  # pragma: no cover

    def __exit__(
        self, exc_type: ExcType, exc_val: ExcValue, exc_tb: ExcTraceback
    ) -> bool:
        """Exit context manager."""
        ...  # pragma: no cover



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import os
import typing

import inotify_simple as _inotify  # type: ignore

DEFAULT_TIMEOUT = 300


class INotify:
    """A waiter that monitors a file path for changes, based on inotify.

    Inotify is used to monitor the specified path for changes (writes).
    It stops waiting when the file is changed or the timeout is reached.

    A `print_func` can be specified as a simple logging method.
    """

    timeout: int = DEFAULT_TIMEOUT
    print_func = None

    def __init__(
        self,
        path: str,
        print_func: typing.Optional[typing.Callable] = None,
        timeout: typing.Optional[int] = None,
    ) -> None:
        if timeout is not None:
            self.timeout = timeout
        self.print_func = print_func
        self._inotify = _inotify.INotify()
        dirpath, fpath = os.path.split(path)
        if not dirpath:
            dirpath = "."
        if not fpath:
            raise ValueError("a file path is required")
        self._dir = dirpath
        self._name = fpath
        self._mask = _inotify.flags.DELETE | _inotify.flags.CLOSE_WRITE
        self._inotify.add_watch(self._dir, self._mask)

    def close(self) -> None:
        self._inotify.close()

    def _print(self, msg: str) -> None:
        if self.print_func:
            self.print_func("[inotify waiter] {}".format(msg))

    def acted(self) -> None:
        return  # noop for inotify waiter

    def wait(self) -> None:
        next(self._wait())

    def _get_events(self) -> list[typing.Any]:
        timeout = 1000 * self.timeout
        self._print("waiting {}ms for activity...".format(timeout))
        events = self._inotify.read(timeout=timeout)
        if not events:
            # use "None" as a sentinel for a timeout, otherwise we can not
            # tell if its all events that didn't match or a true timeout
            return [None]
        # filter out events we don't care about
        return [
            event
            for event in events
            if (event.name == self._name)
            and ((event.mask & _inotify.flags.CLOSE_WRITE) != 0)
        ]

    def _wait(self) -> typing.Iterator[None]:
        while True:
            for event in self._get_events():
                if event is None:
                    self._print("timed out")
                    yield None
                else:
                    self._print(f"{self._name} modified")
                    yield None



#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import errno
import http
import typing
import urllib.error
import urllib.request

from .opener import SchemeNotSupported


class _UnknownHandler(urllib.request.BaseHandler):
    def unknown_open(self, req: urllib.request.Request) -> None:
        raise SchemeNotSupported(req.full_url)


class URLOpener:
    """An Opener type used for fetching remote resources named in
    a pseudo-URL (URI-like) style.
    By default works like urllib.urlopen but only for HTTP(S).

    Example:
    >>> uo = URLOpener()
    >>> res = uo.open("http://abc.example.org/foo/x.json")
    >>> res.read()
    """

    # this list is similar to the defaults found in build_opener
    # but only for http/https handlers. No FTP, etc.
    _handlers = [
        urllib.request.ProxyHandler,
        urllib.request.HTTPHandler,
        urllib.request.HTTPDefaultErrorHandler,
        urllib.request.HTTPRedirectHandler,
        urllib.request.HTTPErrorProcessor,
        urllib.request.HTTPSHandler,
        _UnknownHandler,
    ]

    def __init__(self) -> None:
        self._opener = urllib.request.OpenerDirector()
        for handler in self._handlers:
            self._opener.add_handler(handler())

    def open(self, url: str) -> typing.IO:
        try:
            return self._opener.open(url)
        except ValueError as err:
            # too bad urllib doesn't use a specific subclass of ValueError here
            if "unknown url type" in str(err):
                raise SchemeNotSupported(url) from err
            raise
        except urllib.error.HTTPError as err:
            _map_errno(err)
            raise


_EMAP = {
    http.HTTPStatus.NOT_FOUND.value: errno.ENOENT,
    http.HTTPStatus.UNAUTHORIZED.value: errno.EPERM,
}


def _map_errno(err: urllib.error.HTTPError) -> None:
    """While HTTPError is an OSError, it often doesn't have an errno set.
    Since our callers care about the errno, do a best effort mapping of
    some HTTP statuses to errnos.
    """
    if getattr(err, "errno", None) is not None:
        return
    status = int(getattr(err, "status", -1))
    setattr(err, "errno", _EMAP.get(status, None))



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

from sambacc import samba_cmds, ctdb

from .cli import commands, Context, Fail


def _check_args(parser):
    parser.add_argument(
        "target",
        choices=["winbind", "ctdb-nodestatus"],
        help="Name of the target subsystem to check.",
    )


@commands.command(name="check", arg_func=_check_args)
def check(ctx: Context) -> None:
    """Check that a given subsystem is functioning."""
    if ctx.cli.target == "winbind":
        cmd = samba_cmds.wbinfo["--ping"]
        samba_cmds.execute(cmd)
    elif ctx.cli.target == "ctdb-nodestatus":
        ctdb.check_nodestatus()
    else:
        raise Fail("unknown subsystem: {}".format(ctx.cli.target))



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import typing

from .textfile import TextFileLoader


class NameServiceSwitchLoader(TextFileLoader):
    def __init__(self, path: str) -> None:
        super().__init__(path)
        self.lines: list[str] = []
        self.idx: dict[str, int] = {}

    def loadlines(self, lines: typing.Iterable[str]) -> None:
        """Load in the lines from the text source."""
        # Ignore comments and blank lines
        for line in lines:
            if not line.strip() or line.startswith("#"):
                continue
            self.lines.append(line)
        for lnum, line in enumerate(self.lines):
            if line.startswith("passwd:"):
                self.idx["passwd"] = lnum
            if line.startswith("group:"):
                self.idx["group"] = lnum

    def dumplines(self) -> typing.Iterable[str]:
        """Dump the file content as lines of text."""
        prev = None
        yield "# Generated by sambacc -- DO NOT EDIT\n"
        for line in self.lines:
            if prev and not prev.endswith("\n"):
                yield "\n"
            yield line
            prev = line

    def winbind_enabled(self) -> bool:
        pline = self.lines[self.idx["passwd"]]
        gline = self.lines[self.idx["group"]]
        return ("winbind" in pline) and ("winbind" in gline)

    def ensure_winbind_enabled(self) -> None:
        pidx = self.idx["passwd"]
        if "winbind" not in self.lines[pidx]:
            self.lines[pidx] = "passwd:    files winbind\n"
        gidx = self.idx["group"]
        if "winbind" not in self.lines[gidx]:
            self.lines[gidx] = "group:    files winbind\n"



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import os
import pytest

import sambacc.paths


def test_ensure_samba_dirs_fail(tmp_path):
    # This is missing both "var/lib" and "run"
    with pytest.raises(OSError):
        sambacc.paths.ensure_samba_dirs(root=tmp_path)
    os.mkdir(tmp_path / "var")
    os.mkdir(tmp_path / "var/lib")
    # This is missing "run"
    with pytest.raises(OSError):
        sambacc.paths.ensure_samba_dirs(root=tmp_path)


def test_ensure_samba_dirs_ok(tmp_path):
    os.mkdir(tmp_path / "var")
    os.mkdir(tmp_path / "var/lib")
    os.mkdir(tmp_path / "run")
    sambacc.paths.ensure_samba_dirs(root=tmp_path)


def test_ensure_samba_dirs_already(tmp_path):
    os.mkdir(tmp_path / "var")
    os.mkdir(tmp_path / "var/lib")
    os.mkdir(tmp_path / "var/lib/samba")
    os.mkdir(tmp_path / "var/lib/samba/private")
    os.mkdir(tmp_path / "run")
    os.mkdir(tmp_path / "run/samba/")
    os.mkdir(tmp_path / "run/samba/winbindd")
    sambacc.paths.ensure_samba_dirs(root=tmp_path)


def test_ensure_share_dirs(tmp_path):
    assert not os.path.exists(tmp_path / "foobar")
    sambacc.paths.ensure_share_dirs("foobar", root=str(tmp_path))
    assert os.path.exists(tmp_path / "foobar")

    assert not os.path.exists(tmp_path / "wibble")
    sambacc.paths.ensure_share_dirs("/wibble/cat", root=str(tmp_path))
    assert os.path.exists(tmp_path / "wibble/cat")
    sambacc.paths.ensure_share_dirs("/wibble/cat", root=str(tmp_path))
    assert os.path.exists(tmp_path / "wibble/cat")



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import io
import time

import sambacc.container_dns


J1 = """
{
  "ref": "example",
  "items": [
    {
      "name": "users",
      "ipv4": "192.168.76.40",
      "target": "external"
    },
    {
      "name": "users-cluster",
      "ipv4": "10.235.102.5",
      "target": "internal"
    }
  ]
}
"""

J2 = """
{
  "ref": "example2",
  "items": [
    {
      "name": "users-cluster",
      "ipv4": "10.235.102.5",
      "target": "internal"
    }
  ]
}
"""

J3 = """
{
  "ref": "example",
  "items": [
    {
      "name": "users",
      "ipv4": "192.168.76.108",
      "target": "external"
    },
    {
      "name": "users-cluster",
      "ipv4": "10.235.102.5",
      "target": "internal"
    }
  ]
}
"""


def test_parse():
    fh = io.StringIO(J1)
    hs = sambacc.container_dns.parse(fh)
    assert hs.ref == "example"
    assert len(hs.items) == 2
    assert hs.items[0].name == "users"
    assert hs.items[0].ipv4_addr == "192.168.76.40"


def test_parse2():
    fh = io.StringIO(J2)
    hs = sambacc.container_dns.parse(fh)
    assert hs.ref == "example2"
    assert len(hs.items) == 1
    assert hs.items[0].name == "users-cluster"
    assert hs.items[0].ipv4_addr == "10.235.102.5"


def test_same():
    hs1 = sambacc.container_dns.HostState(ref="apple")
    hs2 = sambacc.container_dns.HostState(ref="orange")
    assert hs1 != hs2
    hs2 = sambacc.container_dns.HostState(ref="apple")
    assert hs1 == hs2
    hs1.items = [
        sambacc.container_dns.HostInfo("a", "10.10.10.10", "external"),
        sambacc.container_dns.HostInfo("b", "10.10.10.11", "external"),
    ]
    hs2.items = [
        sambacc.container_dns.HostInfo("a", "10.10.10.10", "external"),
        sambacc.container_dns.HostInfo("b", "10.10.10.11", "external"),
    ]
    assert hs1 == hs2
    hs2.items = [
        sambacc.container_dns.HostInfo("a", "10.10.10.10", "external"),
        sambacc.container_dns.HostInfo("b", "10.10.10.12", "external"),
    ]
    assert hs1 != hs2
    hs2.items = [
        sambacc.container_dns.HostInfo("a", "10.10.10.10", "external")
    ]
    assert hs1 != hs2


def test_register_dummy(capfd):
    def register(iconfig, hs):
        return sambacc.container_dns.register(
            iconfig,
            hs,
            prefix=["echo"],
        )

    hs = sambacc.container_dns.HostState(
        ref="example",
        items=[
            sambacc.container_dns.HostInfo(
                "foobar", "10.10.10.10", "external"
            ),
            sambacc.container_dns.HostInfo(
                "blat", "192.168.10.10", "internal"
            ),
        ],
    )
    register("example.test", hs)
    out, err = capfd.readouterr()
    assert "net ads -P dns register foobar.example.test 10.10.10.10" in out


def test_parse_and_update(tmp_path):
    reg_data = []

    def _register(domain, hs, target_name=""):
        reg_data.append((domain, hs))
        return True

    path = tmp_path / "test.json"
    with open(path, "w") as fh:
        fh.write(J1)

    hs1, up = sambacc.container_dns.parse_and_update(
        "example.com", path, reg_func=_register
    )
    assert len(reg_data) == 1
    assert up
    hs2, up = sambacc.container_dns.parse_and_update(
        "example.com", path, previous=hs1, reg_func=_register
    )
    assert len(reg_data) == 1
    assert not up

    with open(path, "w") as fh:
        fh.write(J2)
    hs3, up = sambacc.container_dns.parse_and_update(
        "example.com", path, previous=hs2, reg_func=_register
    )
    assert len(reg_data) == 2
    assert reg_data[-1][1] == hs3
    assert up


def test_watch(tmp_path):
    reg_data = []

    def _register(domain, hs, target_name=""):
        reg_data.append((domain, hs))

    def _update(domain, source, previous=None):
        return sambacc.container_dns.parse_and_update(
            domain, source, previous=previous, reg_func=_register
        )

    scount = 0

    def _sleep():
        nonlocal scount
        scount += 1
        if scount > 10:
            raise KeyboardInterrupt()
        time.sleep(0.05)

    path = tmp_path / "test.json"
    with open(path, "w") as fh:
        fh.write(J1)

    sambacc.container_dns.watch(
        "example.com",
        path,
        update_func=_update,
        pause_func=_sleep,
        print_func=lambda x: None,
    )
    assert scount > 10
    assert len(reg_data) == 1

    with open(path, "w") as fh:
        fh.write(J1)
    scount = 0

    def _sleep2():
        nonlocal scount
        scount += 1
        if scount == 5:
            with open(path, "w") as fh:
                fh.write(J3)
        if scount == 10:
            with open(path, "w") as fh:
                fh.write(J3)
        if scount > 20:
            raise KeyboardInterrupt()
        time.sleep(0.05)

    sambacc.container_dns.watch(
        "example.com",
        path,
        update_func=_update,
        pause_func=_sleep2,
        print_func=lambda x: None,
    )
    assert scount > 20
    assert len(reg_data) == 3



#
# sambacc: a samba container configuration tool
# Copyright (C) 2023  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import io

import sambacc.smbconf_api


def test_simple_config_store():
    scs = sambacc.smbconf_api.SimpleConfigStore()
    assert scs.writeable, "SimpleConfigStore should always be writeable"
    scs["foo"] = [("a", "Artichoke"), ("b", "Broccoli")]
    scs["bar"] = [("example", "yes"), ("production", "no")]
    assert list(scs) == ["foo", "bar"]
    assert scs["foo"] == [("a", "Artichoke"), ("b", "Broccoli")]
    assert scs["bar"] == [("example", "yes"), ("production", "no")]


def test_simple_config_store_import():
    a = sambacc.smbconf_api.SimpleConfigStore()
    b = sambacc.smbconf_api.SimpleConfigStore()
    a["foo"] = [("a", "Artichoke"), ("b", "Broccoli")]
    b["bar"] = [("example", "yes"), ("production", "no")]
    assert list(a) == ["foo"]
    assert list(b) == ["bar"]

    a.import_smbconf(b)
    assert list(a) == ["foo", "bar"]
    assert list(b) == ["bar"]
    assert a["bar"] == [("example", "yes"), ("production", "no")]

    b["baz"] = [("quest", "one")]
    b["bar"] = [("example", "no"), ("production", "no"), ("unittest", "yes")]
    a.import_smbconf(b)

    assert list(a) == ["foo", "bar", "baz"]
    assert a["bar"] == [
        ("example", "no"),
        ("production", "no"),
        ("unittest", "yes"),
    ]
    assert a["baz"] == [("quest", "one")]


def test_write_store_as_smb_conf():
    scs = sambacc.smbconf_api.SimpleConfigStore()
    scs["foo"] = [("a", "Artichoke"), ("b", "Broccoli")]
    scs["bar"] = [("example", "yes"), ("production", "no")]
    scs["global"] = [("first", "1"), ("second", "2")]
    fh = io.StringIO()
    sambacc.smbconf_api.write_store_as_smb_conf(fh, scs)
    res = fh.getvalue().splitlines()
    assert res[0] == ""
    assert res[1] == "[global]"
    assert res[2] == "\tfirst = 1"
    assert res[3] == "\tsecond = 2"
    assert "[foo]" in res
    assert "\ta = Artichoke" in res
    assert "\tb = Broccoli" in res
    assert "[bar]" in res
    assert "\texample = yes" in res
    assert "\tproduction = no" in res



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import enum
import logging
import os
import subprocess
import time
import typing

from sambacc import config
from sambacc import leader
from sambacc import samba_cmds
from sambacc.jfile import ClusterMetaJSONFile
from sambacc.netcmd_loader import template_config
from sambacc.typelets import ExcType, ExcValue, ExcTraceback

_logger = logging.getLogger(__name__)


DB_DIR = "/var/lib/ctdb/persistent"
ETC_DIR = "/etc/ctdb"
SHARE_DIR = "/usr/share/ctdb"

CTDB_CONF: str = "/etc/ctdb/ctdb.conf"
CTDB_NODES: str = "/etc/ctdb/nodes"


class ClusterMetaObject(typing.Protocol):
    "A Cluster Meta Object can load or dump persistent cluster descriptions."

    def load(self) -> typing.Any:
        """Load a JSON-compatible object."""
        ...  # pragma: no cover

    def dump(self, data: typing.Any) -> None:
        """Dump (save) a JSON-compatible object."""
        ...  # pragma: no cover


class ClusterMeta(typing.Protocol):
    """ClusterMeta manages access to persistent cluster descriptions."""

    def open(
        self, *, read: bool = True, write: bool = False, locked: bool = False
    ) -> typing.ContextManager[ClusterMetaObject]:
        """Return a context manager with access to a cluster meta object."""
        ...  # pragma: no cover


class NodeState(str, enum.Enum):
    NEW = "new"
    READY = "ready"
    CHANGED = "changed"
    REPLACED = "replaced"
    GONE = "gone"  # reserved


def next_state(state: NodeState) -> NodeState:
    if state == NodeState.NEW:
        return NodeState.READY
    elif state == NodeState.CHANGED:
        return NodeState.REPLACED
    elif state == NodeState.REPLACED:
        return NodeState.READY
    return state


class NodeNotPresent(KeyError):
    def __init__(
        self,
        identity: typing.Any,
        pnn: typing.Optional[typing.Union[str, int]] = None,
    ) -> None:
        super().__init__(identity)
        self.identity = identity
        self.pnn = pnn


def ensure_smb_conf(
    iconfig: config.InstanceConfig, path: str = config.SMB_CONF
) -> None:
    """Ensure that the smb.conf on disk is ctdb and registry enabled."""
    with open(path, "w") as fh:
        write_smb_conf(fh, iconfig)


def write_smb_conf(fh: typing.IO, iconfig: config.InstanceConfig) -> None:
    """Write an smb.conf style output enabling ctdb and samba registry."""
    template_config(fh, iconfig.ctdb_smb_config())


def ensure_ctdb_conf(
    iconfig: config.InstanceConfig, path: str = CTDB_CONF
) -> None:
    """Ensure that the ctdb.conf on disk matches our desired state."""
    with open(path, "w") as fh:
        write_ctdb_conf(fh, iconfig.ctdb_config())


def write_ctdb_conf(
    fh: typing.IO, ctdb_params: dict, enc: typing.Callable = str
) -> None:
    """Write a ctdb.conf style output."""

    def _write_param(name: str, key: str) -> None:
        value = ctdb_params.get(key)
        if value is None:
            return
        fh.write(enc(f"{name} = {value}\n"))

    fh.write(enc("[logging]\n"))
    _write_param("log level", "log_level")
    fh.write(enc("\n"))
    fh.write(enc("[cluster]\n"))
    _write_param("recovery lock", "recovery_lock")
    if ctdb_params.get("nodes_cmd"):
        nodes_cmd = ctdb_params["nodes_cmd"]
        fh.write(enc(f"nodes list = !{nodes_cmd}"))
    fh.write(enc("\n"))
    fh.write(enc("[legacy]\n"))
    _write_param("realtime scheduling", "realtime_scheduling")
    _write_param("script log level", "script_log_level")
    fh.write(enc("\n"))


def ensure_ctdb_nodes(
    ctdb_nodes: list[str], real_path: str, canon_path: str = CTDB_NODES
) -> None:
    """Ensure a real nodes file exists, containing the specificed content,
    and has a symlink in the proper place for ctdb.
    """
    try:
        os.unlink(canon_path)
    except FileNotFoundError:
        pass
    os.symlink(real_path, canon_path)
    # XXX: add locking?
    with open(real_path, "w") as fh:
        write_nodes_file(fh, ctdb_nodes)


def write_nodes_file(
    fh: typing.IO, ctdb_nodes: list[str], enc: typing.Callable = str
) -> None:
    """Write the ctdb nodes file."""
    for node in ctdb_nodes:
        fh.write(enc(f"{node}\n"))


def read_nodes_file(fh: typing.IO) -> list[str]:
    """Read content from an open ctdb nodes file."""
    entries = []
    for line in fh:
        entries.append(line.strip())
    return entries


def read_ctdb_nodes(path: str = CTDB_NODES) -> list[str]:
    """Read the content of the ctdb nodes file."""
    try:
        with open(path, "r") as fh:
            entries = read_nodes_file(fh)
    except FileNotFoundError:
        return []
    return entries


class PublicAddrAssignment(typing.TypedDict):
    address: str
    interfaces: list[str]


def _ensure_public_addresses_file(
    path: str, addrs: list[PublicAddrAssignment]
) -> None:
    with open(path, "w") as fh:
        _write_public_addresses_file(fh, addrs)


def _write_public_addresses_file(
    fh: typing.IO, addrs: list[PublicAddrAssignment]
) -> None:
    for entry in addrs:
        fh.write(entry["address"])
        if entry["interfaces"]:
            ifaces = ",".join(entry["interfaces"])
            fh.write(f" {ifaces}")
        fh.write("\n")


def ensure_ctdb_node_present(
    node: str,
    real_path: str,
    canon_path: str = CTDB_NODES,
    expected_pnn: typing.Optional[int] = None,
) -> None:
    """Ensure that the ctdb nodes file is populated with at least the
    node given. The optional `expect_pnn` can be provided to ensure that
    the node occupies the correct position in the nodes file.
    """
    nodes = read_ctdb_nodes(real_path)
    if node not in nodes:
        nodes.append(node)
    if expected_pnn is not None:
        try:
            found_pnn = nodes.index(node)
        except ValueError:
            found_pnn = -1
        if expected_pnn != found_pnn:
            raise ValueError(f"expected pnn {expected_pnn} is not {found_pnn}")
    ensure_ctdb_nodes(nodes, real_path=real_path, canon_path=canon_path)


def add_node_to_statefile(
    identity: str, node: str, pnn: int, path: str, in_nodes: bool = False
) -> None:
    """Add the given node's identity, (node) IP, and PNN to the JSON based
    state file, located at `path`. If in_nodes is true, the state file will
    reflect that the node is already added to the CTDB nodes file.
    """
    add_node_to_cluster_meta(
        ClusterMetaJSONFile(path), identity, node, pnn, in_nodes=in_nodes
    )


def add_node_to_cluster_meta(
    cmeta: ClusterMeta,
    identity: str,
    node: str,
    pnn: int,
    in_nodes: bool = False,
) -> None:
    """Add the given node's identity, (node) IP, and PNN to the cluster
    metadata.  If in_nodes is true, the state file will reflect that the node
    is already added to the CTDB nodes file.
    """
    with cmeta.open(write=True, locked=True) as cmo:
        data = cmo.load()
        _update_statefile(data, identity, node, pnn, in_nodes=in_nodes)
        cmo.dump(data)


def refresh_node_in_statefile(
    identity: str, node: str, pnn: int, path: str
) -> None:
    """Assuming the node is already in the statefile, update the state in
    the case that the node (IP) has changed.
    """
    refresh_node_in_cluster_meta(
        ClusterMetaJSONFile(path), identity, node, pnn
    )


def refresh_node_in_cluster_meta(
    cmeta: ClusterMeta, identity: str, node: str, pnn: int
) -> None:
    """Assuming the node is already in the cluster metadata, update the state
    in the case that the node (IP) has changed.
    """
    with cmeta.open(write=True, locked=True) as cmo:
        data = cmo.load()
        _refresh_statefile(data, identity, node, pnn)
        cmo.dump(data)


def _update_statefile(
    data: dict[str, typing.Any],
    identity: str,
    node: str,
    pnn: int,
    in_nodes: bool = False,
) -> None:
    data.setdefault("nodes", [])
    for entry in data["nodes"]:
        if pnn == entry["pnn"]:
            raise ValueError("duplicate pnn")
        if identity == entry["identity"]:
            raise ValueError("duplicate identity")
    state = NodeState.NEW
    if in_nodes:
        state = NodeState.READY
    data["nodes"].append(
        {
            "identity": identity,
            "node": node,
            "pnn": pnn,
            "state": state,
        }
    )


def _refresh_statefile(
    data: dict[str, typing.Any],
    identity: str,
    node: str,
    pnn: int,
    in_nodes: bool = False,
) -> None:
    data.setdefault("nodes", [])
    node_entry = None
    for entry in data["nodes"]:
        if pnn == entry["pnn"] and identity == entry["identity"]:
            node_entry = entry
            break
        if pnn == entry["pnn"]:
            raise ValueError(
                f"matching pnn ({pnn}) identity={entry['identity']}"
            )
    if not node_entry:
        raise NodeNotPresent(identity, pnn)
    if node_entry["node"] == node:
        # do nothing
        return
    node_entry["node"] = node
    node_entry["state"] = NodeState.CHANGED


def _get_state(entry: dict[str, typing.Any]) -> NodeState:
    return NodeState(entry["state"])


def _get_state_ok(entry: dict[str, typing.Any]) -> bool:
    return _get_state(entry) == NodeState.READY


def pnn_in_nodes(pnn: int, nodes_json: str, real_path: str) -> bool:
    """Returns true if the specified pnn has an entry in the nodes json
    file and that the node is already added to the ctdb nodes file.
    """
    return pnn_in_cluster_meta(ClusterMetaJSONFile(nodes_json), pnn)


def pnn_in_cluster_meta(cmeta: ClusterMeta, pnn: int) -> bool:
    """Returns true if the specified pnn has an entry in the cluster metadata
    and that entry is ready for use.
    """
    with cmeta.open(locked=True) as cmo:
        json_data = cmo.load()
    current_nodes = json_data.get("nodes", [])
    for entry in current_nodes:
        if pnn == entry["pnn"] and _get_state_ok(entry):
            return True
    return False


def manage_nodes(
    pnn: int,
    nodes_json: str,
    real_path: str,
    pause_func: typing.Callable,
) -> None:
    """Monitor nodes json for updates, reflecting those changes into ctdb."""
    manage_cluster_meta_updates(
        ClusterMetaJSONFile(nodes_json),
        pnn,
        real_path,
        pause_func,
    )


def manage_cluster_meta_updates(
    cmeta: ClusterMeta,
    pnn: int,
    real_path: str,
    pause_func: typing.Callable,
) -> None:
    """Monitor cluster meta for updates, reflecting those changes into ctdb."""
    while True:
        _logger.info("checking if node is able to make updates")
        if _node_check(cmeta, pnn, real_path):
            _logger.info("checking for node updates")
            if _node_update(cmeta, real_path):
                _logger.info("updated nodes")
        else:
            _logger.warning("node can not make updates")
        pause_func()


def _node_check(cmeta: ClusterMeta, pnn: int, real_path: str) -> bool:
    with cmeta.open(locked=True) as cmo:
        desired = cmo.load().get("nodes", [])
    ctdb_nodes = read_ctdb_nodes(real_path)
    # first: check to see if the current node is in the nodes file
    try:
        my_desired = [e for e in desired if e.get("pnn") == pnn][0]
    except IndexError:
        # no entry found for this node
        _logger.warning(f"PNN {pnn} not found in json state file")
        return False
    if my_desired["node"] not in ctdb_nodes:
        # this current node is not in the nodes file.
        # it is ineligible to make changes to the nodes file
        return False
    # this node is already in the nodes file!
    return True


def _node_update_check(
    json_data: dict[str, typing.Any], real_path: str
) -> tuple[list[str], list[typing.Any], list[typing.Any]]:
    desired = json_data.get("nodes", [])
    ctdb_nodes = read_ctdb_nodes(real_path)
    update_nodes = []
    need_reload = []
    _update_states = (NodeState.NEW, NodeState.CHANGED, NodeState.REPLACED)
    for entry in desired:
        pnn = entry["pnn"]
        matched = _node_line(ctdb_nodes, pnn) == entry["node"]
        if matched and _get_state_ok(entry):
            # everything's fine. skip this entry
            continue
        elif not matched:
            if entry["state"] in _update_states:
                update_nodes.append(entry)
                need_reload.append(entry)
            elif entry["state"] == NodeState.READY:
                msg = f"ready node (pnn {pnn}) missing from {ctdb_nodes}"
                raise ValueError(msg)
        else:
            # node present but state indicates
            # update is not finalized
            need_reload.append(entry)
    return ctdb_nodes, update_nodes, need_reload


def _node_line(ctdb_nodes: list[str], pnn: int) -> str:
    try:
        return ctdb_nodes[pnn]
    except IndexError:
        return ""


def _entry_to_node(ctdb_nodes: list[str], entry: dict[str, typing.Any]) -> str:
    pnn: int = entry["pnn"]
    if entry["state"] == NodeState.CHANGED:
        return "#{}".format(ctdb_nodes[pnn].strip("#"))
    return entry["node"]


def _node_update(cmeta: ClusterMeta, real_path: str) -> bool:
    # open r/o so that we don't initailly open for write.  we do a probe and
    # decide if anything needs to be updated if we are wrong, its not a
    # problem, we'll "time out" and reprobe later
    with cmeta.open(locked=True) as cmo:
        json_data = cmo.load()
        _, test_chg_nodes, test_need_reload = _node_update_check(
            json_data, real_path
        )
        if not test_chg_nodes and not test_need_reload:
            _logger.info("examined nodes state - no changes")
            return False
    # we probably need to make a change. but we recheck our state again
    # under lock, with the data file open r/w
    # update the nodes file and make changes to ctdb
    with cmeta.open(write=True, locked=True) as cmo:
        json_data = cmo.load()
        ctdb_nodes, chg_nodes, need_reload = _node_update_check(
            json_data, real_path
        )
        if not chg_nodes and not need_reload:
            _logger.info("reexamined nodes state - no changes")
            return False
        _logger.info("writing updates to ctdb nodes file")
        new_ctdb_nodes = list(ctdb_nodes)
        for entry in chg_nodes:
            pnn = entry["pnn"]
            expected_line = _entry_to_node(ctdb_nodes, entry)
            if _node_line(new_ctdb_nodes, pnn) == expected_line:
                continue
            if entry["state"] == NodeState.NEW:
                if pnn != len(new_ctdb_nodes):
                    raise ValueError(
                        f"unexpected pnn in new entry {entry}:"
                        " nodes: {new_ctdb_nodes}"
                    )
                new_ctdb_nodes.append(expected_line)
            else:
                new_ctdb_nodes[pnn] = expected_line
        _save_nodes(real_path, new_ctdb_nodes)
        _logger.info("running: ctdb reloadnodes")
        subprocess.check_call(list(samba_cmds.ctdb["reloadnodes"]))
        for entry in need_reload:
            entry["state"] = next_state(entry["state"])
            _logger.debug(
                "setting node identity=[{}] pnn={} to {}".format(
                    entry["identity"],
                    entry["pnn"],
                    entry["state"],
                )
            )
        cmo.dump(json_data)
    return True


def cluster_meta_to_nodes(
    cmeta: ClusterMeta, dest: typing.Union[str, typing.IO]
) -> None:
    """Write a nodes file based on the current content of the cluster
    metadata."""
    with cmeta.open(locked=True) as cmo:
        json_data = cmo.load()
        nodes = json_data.get("nodes", [])
        _logger.info("Found node metadata: %r", nodes)
        ctdb_nodes = _cluster_meta_to_ctdb_nodes(nodes)
        if isinstance(dest, str):
            _logger.info("Will write nodes: %s", ctdb_nodes)
            _save_nodes(dest, ctdb_nodes)
        else:
            write_nodes_file(dest, ctdb_nodes)


def _cluster_meta_to_ctdb_nodes(nodes: list[dict]) -> list[str]:
    pnn_max = max(n["pnn"] for n in nodes) + 1  # pnn is zero indexed
    ctdb_nodes: list[str] = [""] * pnn_max
    for entry in nodes:
        pnn = entry["pnn"]
        # overwrite the pnn indexed entry with expected value
        ctdb_nodes[pnn] = _entry_to_node(ctdb_nodes, entry)
    return ctdb_nodes


def _save_nodes(path: str, ctdb_nodes: list[str]) -> None:
    with open(path, "w") as nffh:
        write_nodes_file(nffh, ctdb_nodes)
        nffh.flush()
        os.fsync(nffh)


def monitor_cluster_meta_changes(
    cmeta: ClusterMeta,
    pause_func: typing.Callable,
    *,
    nodes_file_path: typing.Optional[str] = None,
    reload_all: bool = False,
    leader_locator: typing.Optional[leader.LeaderLocator] = None,
) -> None:
    """Monitor cluster meta for changes, reflecting those changes into ctdb.

    Unlike manage_cluster_meta_updates this function never changes the
    contents of the nodes list in the cluster meta and takes those values
    as a given, assuming some external agent has the correct global view of
    the cluster and is updating it correctly. This function exists to
    translate that content into something ctdb can understand.
    """
    prev_meta: dict[str, typing.Any] = {}
    if nodes_file_path:
        prev_nodes = read_ctdb_nodes(nodes_file_path)
    else:
        with cmeta.open(locked=True) as cmo:
            meta1 = cmo.load()
        prev_nodes = _cluster_meta_to_ctdb_nodes(meta1.get("nodes", []))
    _logger.debug("initial cluster meta content: %r", prev_meta)
    _logger.debug("initial nodes content: %r", prev_nodes)
    while True:
        pause_func()
        with cmeta.open(locked=True) as cmo:
            curr_meta = cmo.load()
        if curr_meta == prev_meta:
            _logger.debug("cluster meta content unchanged: %r", curr_meta)
            continue
        if len(prev_meta) > 0 and len(curr_meta) == 0:
            # cluster is possibly (probably?) being destroyed.
            # Return from this loop and let the command-level loop decide if
            # this function needs to be restarted or not. There's a chance this
            # process will be terminated very soon anyway.
            _logger.warning("no current nodes available")
            return
        _logger.info("cluster meta content changed")
        _logger.debug(
            "cluster meta: previous=%r current=%r", prev_meta, curr_meta
        )
        prev_meta = curr_meta

        # maybe some other metadata changed?
        expected_nodes = _cluster_meta_to_ctdb_nodes(
            curr_meta.get("nodes", [])
        )
        if prev_nodes == expected_nodes:
            _logger.debug("ctdb nodes list unchanged: %r", expected_nodes)
            continue
        _logger.info("ctdb nodes list changed")
        _logger.debug(
            "nodes list: previous=%r current=%r", prev_nodes, expected_nodes
        )
        prev_nodes = expected_nodes

        if nodes_file_path:
            _logger.info("updating nodes file: %s", nodes_file_path)
            _save_nodes(nodes_file_path, expected_nodes)
        _maybe_reload_nodes_retry(leader_locator, reload_all=reload_all)


def _maybe_reload_nodes_retry(
    leader_locator: typing.Optional[leader.LeaderLocator] = None,
    reload_all: bool = False,
    *,
    tries: int = 5,
) -> None:
    for idx in range(tries):
        time.sleep(1 << idx)
        try:
            _maybe_reload_nodes(leader_locator, reload_all=reload_all)
            return
        except subprocess.CalledProcessError:
            _logger.exception("failed to execute reload nodes command")
    raise RuntimeError("exceeded retries running reload nodes command")


def _maybe_reload_nodes(
    leader_locator: typing.Optional[leader.LeaderLocator] = None,
    reload_all: bool = False,
) -> None:
    """Issue a reloadnodes command if leader_locator is available and
    node is leader or reload_all is true.
    """
    if reload_all:
        _logger.info("running: ctdb reloadnodes")
        subprocess.check_call(list(samba_cmds.ctdb["reloadnodes"]))
        return
    if leader_locator is None:
        _logger.warning("no leader locator: not calling reloadnodes")
        return
    # use the leader locator to only issue the reloadnodes command once
    # for a change instead of all the nodes "spamming" the cluster
    with leader_locator as ll:
        if ll.is_leader():
            _logger.info("running: ctdb reloadnodes")
            subprocess.check_call(list(samba_cmds.ctdb["reloadnodes"]))
        else:
            _logger.info("node is not leader. skipping reloadnodes")


def ensure_ctdbd_etc_files(
    etc_path: str = ETC_DIR,
    src_path: str = SHARE_DIR,
    *,
    iconfig: typing.Optional[config.InstanceConfig] = None,
) -> None:
    """Ensure certain files that ctdbd expects to exist in its etc dir
    do exist.
    """
    functions_src = os.path.join(src_path, "functions")
    functions_dst = os.path.join(etc_path, "functions")
    notify_src = os.path.join(src_path, "notify.sh")
    notify_dst = os.path.join(etc_path, "notify.sh")
    legacy_scripts_src = os.path.join(src_path, "events/legacy")
    legacy_scripts_dst = os.path.join(etc_path, "events/legacy")
    link_legacy_scripts = ["00.ctdb.script"]

    public_addresses: list[PublicAddrAssignment] = []
    if iconfig:
        ctdb_conf = iconfig.ctdb_config()
        # todo: when we have a real config object for ctdb conf we can drop
        # the typing.cast
        public_addresses = typing.cast(
            list[PublicAddrAssignment], ctdb_conf.get("public_addresses", [])
        )
    if public_addresses:
        link_legacy_scripts.append("10.interface.script")

    os.makedirs(etc_path, exist_ok=True)
    try:
        os.unlink(functions_dst)
    except FileNotFoundError:
        pass
    os.symlink(functions_src, functions_dst)

    try:
        os.unlink(notify_dst)
    except FileNotFoundError:
        pass
    os.symlink(notify_src, notify_dst)

    os.makedirs(legacy_scripts_dst, exist_ok=True)
    for legacy_script_name in link_legacy_scripts:
        lscript_src = os.path.join(legacy_scripts_src, legacy_script_name)
        lscript_dst = os.path.join(legacy_scripts_dst, legacy_script_name)
        try:
            os.unlink(lscript_dst)
        except FileNotFoundError:
            pass
        os.symlink(lscript_src, lscript_dst)

    if public_addresses:
        pa_path = os.path.join(etc_path, "public_addresses")
        _ensure_public_addresses_file(pa_path, public_addresses)


_SRC_TDB_FILES = [
    "account_policy.tdb",
    "group_mapping.tdb",
    "passdb.tdb",
    "registry.tdb",
    "secrets.tdb",
    "share_info.td",
    "winbindd_idmap.tdb",
]

_SRC_TDB_DIRS = [
    "/var/lib/samba",
    "/var/lib/samba/private",
]


def migrate_tdb(
    iconfig: config.InstanceConfig, dest_dir: str, pnn: int = 0
) -> None:
    """Migrate TDB files into CTDB."""
    # TODO: these paths should be based on our instance config, not hard coded
    for tdbfile in _SRC_TDB_FILES:
        for parent in _SRC_TDB_DIRS:
            tdb_path = os.path.join(parent, tdbfile)
            if _has_tdb_file(tdb_path):
                _convert_tdb_file(tdb_path, dest_dir, pnn=pnn)


def _has_tdb_file(tdb_path: str) -> bool:
    # TODO: It would be preferable to handle errors from the convert
    # function only, but it if ltdbtool is missing it raises FileNotFoundError
    # and its not simple to disambiguate between the command missing and the
    # tdb file missing.
    _logger.info(f"Checking for {tdb_path}")
    return os.path.isfile(tdb_path)


def _convert_tdb_file(tdb_path: str, dest_dir: str, pnn: int = 0) -> None:
    orig_name = os.path.basename(tdb_path)
    opath = os.path.join(dest_dir, f"{orig_name}.{pnn}")
    _logger.info(f"Converting {tdb_path} to {opath} ...")
    cmd = samba_cmds.ltdbtool["convert", "-s0", tdb_path, opath]
    subprocess.check_call(list(cmd))


def archive_tdb(iconfig: config.InstanceConfig, dest_dir: str) -> None:
    """Arhive TDB files into a given directory."""
    # TODO: these paths should be based on our instance config, not hard coded
    try:
        os.mkdir(dest_dir)
        _logger.debug("dest_dir: %r created", dest_dir)
    except FileExistsError:
        _logger.debug("dest_dir: %r already exists", dest_dir)
    for tdbfile in _SRC_TDB_FILES:
        for parent in _SRC_TDB_DIRS:
            tdb_path = os.path.join(parent, tdbfile)
            if _has_tdb_file(tdb_path):
                dest_path = os.path.join(dest_dir, tdbfile)
                _logger.info("archiving: %r -> %r", tdb_path, dest_path)
                os.rename(tdb_path, dest_path)


def check_nodestatus(cmd: samba_cmds.SambaCommand = samba_cmds.ctdb) -> None:
    cmd_ctdb_check = cmd["nodestatus"]
    samba_cmds.execute(cmd_ctdb_check)


def _read_command_pnn(cmd: samba_cmds.SambaCommand) -> typing.Optional[int]:
    """Run a ctdb command assuming it returns a pnn value. Return the pnn as an
    int on success, None on command failure.
    """
    try:
        out = subprocess.check_output(list(cmd))
        pnntxt = out.decode("utf8").strip()
    except subprocess.CalledProcessError as err:
        _logger.error(f"command {cmd!r} failed: {err!r}")
        return None
    except FileNotFoundError:
        _logger.error(f"ctdb command ({cmd!r}) not found")
        return None
    try:
        return int(pnntxt)
    except ValueError:
        _logger.debug(f"ctdb command wrote invalid pnn: {pnntxt!r}")
        return None


def current_pnn() -> typing.Optional[int]:
    """Run the `ctdb pnn` command. Returns the pnn value or None if the command
    fails.
    """
    return _read_command_pnn(samba_cmds.ctdb["pnn"])


def leader_pnn() -> typing.Optional[int]:
    """Run the `ctdb leader` (or equivalent) command. Returns the pnn value or
    None if the command fails.
    """
    # recmaster command: <ctdb recmaster|leader>
    admin_cmd = samba_cmds.ctdb_leader_admin_cmd()
    return _read_command_pnn(samba_cmds.ctdb[admin_cmd])


class CLILeaderStatus:
    _isleader = False

    def is_leader(self) -> bool:
        return self._isleader


class CLILeaderLocator:
    """A leader locator that relies entirely on checking the
    recovery master using the ctdb command line tool.
    """

    def __enter__(self) -> CLILeaderStatus:
        mypnn = current_pnn()
        leader = leader_pnn()
        sts = CLILeaderStatus()
        sts._isleader = mypnn is not None and mypnn == leader
        return sts

    def __exit__(
        self, exc_type: ExcType, exc_val: ExcValue, exc_tb: ExcTraceback
    ) -> bool:
        return True



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import argparse
import contextlib
import logging
import os
import socket
import sys
import typing

from sambacc import ctdb
from sambacc import jfile
from sambacc import rados_opener
from sambacc import samba_cmds
from sambacc.simple_waiter import Sleeper, Waiter

from .cli import best_leader_locator, best_waiter, commands, Context, Fail

_logger = logging.getLogger(__name__)

# Rather irritatingly, k8s does not have a simple method for passing the
# ordinal index of a stateful set down to the containers. This has been
# proposed but not implemented yet. See:
#  https://github.com/kubernetes/kubernetes/issues/40651
# While I find putting any k8s specific knowledge in sambacc distasteful
# all we're really doing is teaching sambacc how to extract the node
# number from the host name, an operation that's not k8s specific.
# That isn't *too* dirty. Just a smudge really. :-)
_AFTER_LAST_DASH = "after-last-dash"


def _ctdb_ok():
    sambacc_ctdb = os.environ.get("SAMBACC_CTDB")
    gate = "ctdb-is-experimental"
    if sambacc_ctdb == gate:
        return
    print("Using CTDB with samba-container (sambacc) is experimental.")
    print("If you are developing or testing features for sambacc please")
    print("set the environment variable SAMBACC_CTDB to the value:")
    print("    ", gate)
    print("before continuing and try again.")
    print()
    raise Fail(gate)


def _ctdb_migrate_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--dest-dir",
        default=ctdb.DB_DIR,
        help="Specify where CTDB database files will be written.",
    )
    parser.add_argument(
        "--archive",
        help="Move converted TDB files to an archive dir.",
    )


def _ctdb_general_node_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--hostname",
        help="Specify the host name for the CTDB node",
    )
    parser.add_argument(
        "--node-number",
        type=int,
        help="Expected node number",
    )
    # This is a choice with a single acceptable param, rather than an on/off
    # bool, # in the case that other container orchs have a similar but not
    # quite the same issue and we want to support a different scheme someday.
    parser.add_argument(
        "--take-node-number-from-hostname",
        choices=(_AFTER_LAST_DASH,),
        help=(
            "Take the node number from the given host name following"
            " the specified policy."
        ),
    )
    parser.add_argument(
        "--take-node-number-from-env",
        "-E",
        const="NODE_NUMBER",
        nargs="?",
        help=(
            "Take the node number from the environment. If specified"
            " with a value, use that value as the environment variable"
            " name. Otherwise, use environment variable NODE_NUMBER."
        ),
    )
    parser.add_argument(
        "--persistent-path",
        help="Path to a persistent path for storing nodes file",
    )
    parser.add_argument(
        "--metadata-source",
        help=(
            "Specify location of cluster metadata state-tracking object."
            " This can be a file path or a URI-style identifier."
        ),
    )


def _ctdb_set_node_args(parser: argparse.ArgumentParser) -> None:
    _ctdb_general_node_args(parser)
    parser.add_argument(
        "--ip",
        help="Specify node by IP",
    )


class NodeParams:
    _ctx: Context
    node_number: typing.Optional[int] = None
    hostname: typing.Optional[str] = None
    persistent_path: str = ""
    _nodes_json: str = ""
    _cluster_meta_uri: str = ""
    _ip_addr: typing.Optional[str] = None
    _cluster_meta_obj: typing.Optional[ctdb.ClusterMeta] = None
    _waiter_obj: typing.Optional[Waiter] = None

    def __init__(self, ctx: Context):
        self._ctx = ctx
        ccfg = ctx.instance_config.ctdb_config()

        # stuff that many of the commands use
        self.persistent_path = ctx.cli.persistent_path
        if self.persistent_path is None:
            self.persistent_path = ccfg["nodes_path"]
        # nodes_json will now only be in the ctdb config section if it has been
        # specified by the user.
        self._nodes_json = ccfg.get("nodes_json") or ""
        # cluster_meta_uri can be a uri-ish string or path. It will be set with
        # a default value by the config even if there's no user supplied value.
        self._cluster_meta_uri = ccfg.get("cluster_meta_uri") or ""

        self.hostname = ctx.cli.hostname
        if ctx.cli.node_number is not None:
            if ctx.cli.node_number < 0:
                raise ValueError(f"invalid node number: {ctx.cli.node_number}")
            self.node_number = ctx.cli.node_number
        elif ctx.cli.take_node_number_from_hostname == _AFTER_LAST_DASH:
            if not self.hostname:
                raise ValueError(
                    "--hostname required if taking node number from host name"
                )
            if "-" not in self.hostname:
                raise ValueError(
                    f"invalid hostname for node number: {self.hostname}"
                )
            self.node_number = int(self.hostname.rsplit("-")[-1])
        elif ctx.cli.take_node_number_from_env:
            try:
                self.node_number = int(
                    os.environ[ctx.cli.take_node_number_from_env]
                )
            except (KeyError, ValueError):
                raise ValueError(
                    "failed to get node number from environment var"
                    f" {ctx.cli.take_node_number_from_env}"
                )
        else:
            self.node_number = None

    @property
    def node_ip_addr(self) -> str:
        if self._ip_addr is None:
            cli = self._ctx.cli
            if getattr(cli, "ip", None):
                self._ip_addr = cli.ip
            elif cli.hostname:
                self._ip_addr = _lookup_hostname(cli.hostname)
            else:
                raise ValueError("can not determine node ip")
        return self._ip_addr

    @property
    def identity(self) -> str:
        # this could be extended to use something like /etc/machine-id
        # or whatever in the future.
        if self.hostname:
            return self.hostname
        elif self.node_number:
            return f"node-{self.node_number}"
        else:
            # the dashes make this an invalid dns name
            return "-unknown-"

    @property
    def cluster_meta_uri(self) -> str:
        """Return a cluster meta uri value."""
        values = (
            # cli takes highest precedence
            self._ctx.cli.metadata_source,
            # _nodes_json should only be set if user set it using the old key
            self._nodes_json,
            # default or customized value on current key
            self._cluster_meta_uri,
        )
        for uri in values:
            if uri:
                return uri
        raise ValueError("failed to determine cluster_meta_uri")

    def _cluster_meta_init(self) -> None:
        uri = self.cluster_meta_uri
        # it'd be nice to re-use the opener infrastructure here but openers
        # don't do file modes the way we need for JSON state file or do
        # writable file types in the url_opener (urllib wrapper). For now, just
        # manually handle the string.
        if rados_opener.is_rados_uri(uri):
            self._cluster_meta_obj = (
                rados_opener.ClusterMetaRADOSObject.create_from_uri(uri)
            )
            self._waiter_obj = Sleeper()
            return
        if uri.startswith("file:"):
            path = uri.split(":", 1)[-1]
        else:
            path = uri
        if path.startswith("/"):
            path = "/" + path.rstrip("/")  # ensure one leading /
        self._cluster_meta_obj = jfile.ClusterMetaJSONFile(path)
        self._waiter_obj = best_waiter(path)

    def cluster_meta(self) -> ctdb.ClusterMeta:
        if self._cluster_meta_obj is None:
            self._cluster_meta_init()
        assert self._cluster_meta_obj is not None
        return self._cluster_meta_obj

    def cluster_meta_waiter(self) -> Waiter:
        if self._waiter_obj is None:
            self._cluster_meta_init()
        assert self._waiter_obj is not None
        return self._waiter_obj


@commands.command(name="ctdb-migrate", arg_func=_ctdb_migrate_args)
def ctdb_migrate(ctx: Context) -> None:
    """Migrate standard samba databases to CTDB databases."""
    _ctdb_ok()
    ctdb.migrate_tdb(ctx.instance_config, ctx.cli.dest_dir)
    if ctx.cli.archive:
        ctdb.archive_tdb(ctx.instance_config, ctx.cli.archive)


def _lookup_hostname(hostname: str) -> str:
    try:
        addrinfo = socket.getaddrinfo(
            hostname,
            None,
            family=socket.AF_UNSPEC,
            type=socket.SOCK_STREAM,
        )
        ipv6_address = None

        for entry in addrinfo:
            family, _, _, _, sockaddr = entry
            ip_address = sockaddr[0]

            if ip_address.startswith("127.") or ip_address == "::1":
                continue

            if family == socket.AF_INET:
                return ip_address

            if family == socket.AF_INET6 and ipv6_address is None:
                ipv6_address = ip_address

        if ipv6_address:
            return ipv6_address

        raise RuntimeError(
            f"No valid IP address found for hostname '{hostname}'."
        )

    except socket.gaierror as e:
        _logger.error(f"Failed to resolve hostname '{hostname}': {e}")
        raise


@commands.command(name="ctdb-set-node", arg_func=_ctdb_set_node_args)
def ctdb_set_node(ctx: Context) -> None:
    """Set up the current node in the ctdb and sambacc nodes files."""
    _ctdb_ok()
    np = NodeParams(ctx)
    expected_pnn = np.node_number

    try:
        ctdb.refresh_node_in_cluster_meta(
            cmeta=np.cluster_meta(),
            identity=np.identity,
            node=np.node_ip_addr,
            pnn=int(expected_pnn or 0),
        )
        return
    except ctdb.NodeNotPresent:
        pass

    ctdb.add_node_to_cluster_meta(
        cmeta=np.cluster_meta(),
        identity=np.identity,
        node=np.node_ip_addr,
        pnn=int(expected_pnn or 0),
        in_nodes=(expected_pnn == 0),
    )
    if expected_pnn == 0:
        ctdb.ensure_ctdb_node_present(
            node=np.node_ip_addr,
            expected_pnn=expected_pnn,
            real_path=np.persistent_path,
        )


@commands.command(name="ctdb-manage-nodes", arg_func=_ctdb_general_node_args)
def ctdb_manage_nodes(ctx: Context) -> None:
    """Run a long lived process to manage the cluster metadata. It can add new
    nodes. When a new node is found, if the current node is in the correct
    state, this node will add it to CTDB.
    """
    _ctdb_ok()
    np = NodeParams(ctx)
    expected_pnn = np.node_number or 0
    waiter = np.cluster_meta_waiter()

    limiter = ErrorLimiter("ctdb_manage_nodes", 10, pause_func=waiter.wait)
    while True:
        with limiter.catch():
            ctdb.manage_cluster_meta_updates(
                cmeta=np.cluster_meta(),
                pnn=expected_pnn,
                real_path=np.persistent_path,
                pause_func=waiter.wait,
            )


def _ctdb_monitor_nodes_args(parser: argparse.ArgumentParser) -> None:
    _ctdb_must_have_node_args(parser)
    parser.add_argument(
        "--reload",
        choices=("leader", "never", "all"),
        default="leader",
        help="Specify which nodes can command CTDB to reload nodes",
    )


@commands.command(name="ctdb-monitor-nodes", arg_func=_ctdb_monitor_nodes_args)
def ctdb_monitor_nodes(ctx: Context) -> None:
    """Run a long lived process to monitor the cluster metadata.
    Unlike ctdb_manage_nodes this function assumes that the node state
    file is externally managed and primarily exists to reflect any changes
    to the cluster meta into CTDB.
    """
    _ctdb_ok()
    np = NodeParams(ctx)
    waiter = np.cluster_meta_waiter()
    leader_locator = None
    if ctx.cli.reload == "leader":
        leader_locator = best_leader_locator(ctx.instance_config)
    reload_all = ctx.cli.reload == "all"
    nodes_file_path = np.persistent_path if ctx.cli.write_nodes else None

    _logger.info("monitoring cluster meta changes")
    _logger.debug(
        "reload_all=%s leader_locator=%r", reload_all, leader_locator
    )
    limiter = ErrorLimiter("ctdb_monitor_nodes", 10, pause_func=waiter.wait)
    while True:
        with limiter.catch():
            ctdb.monitor_cluster_meta_changes(
                cmeta=np.cluster_meta(),
                pause_func=waiter.wait,
                nodes_file_path=nodes_file_path,
                leader_locator=leader_locator,
                reload_all=reload_all,
            )


def _ctdb_must_have_node_args(parser: argparse.ArgumentParser) -> None:
    _ctdb_general_node_args(parser)
    parser.add_argument(
        "--write-nodes",
        action="store_true",
        help="Write ctdb nodes file based on cluster meta contents",
    )


@commands.command(
    name="ctdb-must-have-node", arg_func=_ctdb_must_have_node_args
)
def ctdb_must_have_node(ctx: Context) -> None:
    """Block until the current node is present in the ctdb nodes file."""
    _ctdb_ok()
    np = NodeParams(ctx)
    expected_pnn = np.node_number or 0
    waiter = np.cluster_meta_waiter()

    limiter = ErrorLimiter("ctdb_must_have_node", 10, pause_func=waiter.wait)
    while True:
        with limiter.catch():
            if ctdb.pnn_in_cluster_meta(
                cmeta=np.cluster_meta(),
                pnn=expected_pnn,
            ):
                break
            _logger.info("node not yet ready")
            waiter.wait()
    if ctx.cli.write_nodes:
        _logger.info("Writing nodes file")
        ctdb.cluster_meta_to_nodes(np.cluster_meta(), dest=np.persistent_path)


def _ctdb_rados_mutex_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--cluster-name",
        default="ceph",
        help="Cluster name to pass to mutex lock helper",
    )
    parser.add_argument(
        "mutex_uri",
        help="RADOS (pesudo) URI value for the object to use as a mutex",
    )


@commands.command(name="ctdb-rados-mutex", arg_func=_ctdb_rados_mutex_args)
def ctdb_rados_mutex(ctx: Context) -> None:
    """A command to wrap the rados ctdb_mutex_ceph_rados_helper and wrap
    & translate the container's ceph configuration into something
    the helper can understand.
    N.B. Another reason for this command is that ctdb requires the
    `cluster lock` value to be the same on all nodes.
    """
    if not rados_opener.is_rados_uri(ctx.cli.mutex_uri):
        raise ValueError(f"{ctx.cli.mutex_uri} is not a valid RADOS URI value")
    rinfo = rados_opener.parse_rados_uri(ctx.cli.mutex_uri)
    if rinfo["subtype"] != "object":
        raise ValueError(
            f"{ctx.cli.mutex_uri} is not a RADOS object URI value"
        )
    pool, namespace, objname = rinfo["pool"], rinfo["ns"], rinfo["key"]
    entity = ctx.cli.ceph_id["client_name"]
    if not entity:
        raise ValueError("a ceph authentication entity name is required")
    if not ctx.cli.ceph_id["full_name"]:
        entity = f"client.{entity}"
    # required arguments
    cmd = samba_cmds.ctdb_mutex_ceph_rados_helper[
        ctx.cli.cluster_name, entity, pool, objname  # cephx entity
    ]
    # optional namespace argument
    if namespace:
        cmd = cmd["-n", namespace]
    _logger.debug("executing command: %r", cmd)
    samba_cmds.execute(cmd)  # replaces process


@commands.command(name="ctdb-list-nodes", arg_func=_ctdb_general_node_args)
def ctdb_list_nodes(ctx: Context) -> None:
    """Write nodes content to stdout based on current cluster meta."""
    _ctdb_ok()
    np = NodeParams(ctx)

    ctdb.cluster_meta_to_nodes(np.cluster_meta(), sys.stdout)


class ErrorLimiter:
    def __init__(
        self,
        name: str,
        limit: int,
        *,
        pause_func: typing.Optional[typing.Callable] = None,
    ) -> None:
        self.name = name
        self.limit = limit
        self.errors = 0
        self.pause_func = pause_func

    def post_catch(self):
        if self.pause_func is not None:
            self.pause_func()

    @contextlib.contextmanager
    def catch(self) -> typing.Iterator[None]:
        try:
            _logger.debug(
                "error limiter proceeding: %s: errors=%r",
                self.name,
                self.errors,
            )
            yield
        except KeyboardInterrupt:
            raise
        except Exception as err:
            _logger.error(
                f"error during {self.name}: {err}, count={self.errors}",
                exc_info=True,
            )
            self.errors += 1
            if self.errors > self.limit:
                _logger.error(f"too many retries ({self.errors}). giving up")
                raise
            self.post_catch()



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import sambacc.passdb_loader as passdb
import sambacc.passwd_loader as ugl

from .cli import commands, setup_steps, Context


@commands.command(name="import-users")
def import_users(ctx: Context) -> None:
    """Import users and groups from the sambacc config to the passwd
    and group files to support local (non-domain based) login.
    """
    import_sys_users(ctx)
    import_passdb_users(ctx)


@setup_steps.command("users")
def import_sys_users(ctx: Context) -> None:
    """Import users and groups from sambacc config to the passwd and
    group files.
    """
    etc_passwd_loader = ugl.PasswdFileLoader(ctx.cli.etc_passwd_path)
    etc_group_loader = ugl.GroupFileLoader(ctx.cli.etc_group_path)

    etc_passwd_loader.read()
    etc_group_loader.read()
    for u in ctx.instance_config.users():
        etc_passwd_loader.add_user(u)
    for g in ctx.instance_config.groups():
        etc_group_loader.add_group(g)
    etc_passwd_loader.write()
    etc_group_loader.write()


@setup_steps.command("users_passdb")
def import_passdb_users(ctx: Context) -> None:
    """Import users into samba's passdb."""
    smb_passdb_loader = passdb.PassDBLoader()
    for u in ctx.instance_config.users():
        smb_passdb_loader.add_user(u)
    return



#
# sambacc: a samba container configuration tool
# Copyright (C) 2022  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import argparse
import functools
import os

import sambacc.config
import sambacc.opener
import sambacc.paths

import sambacc.commands.config


config1 = """
{
    "samba-container-config": "v0",
    "configs": {
        "updateme":{
          "shares": ["uno"],
          "globals": ["global0"]
        }
    },
    "shares": {
       "uno": {
          "options": {
              "path": "/srv/data/uno"
          }
       }
    },
    "globals": {
        "global0": {
           "options": {
             "server min protocol": "SMB2"
           }
        }
    }
}
"""

config2 = """
{
    "samba-container-config": "v0",
    "configs": {
        "updateme":{
          "shares": ["uno", "dos"],
          "globals": ["global0"],
          "permissions": {"method": "none"}
        }
    },
    "shares": {
       "uno": {
          "options": {
              "path": "/srv/data/uno"
          }
       },
       "dos": {
          "options": {
              "path": "/srv/data/dos"
          }
       }
    },
    "globals": {
        "global0": {
           "options": {
             "server min protocol": "SMB2"
           }
        }
    }
}
"""


class FakeContext:
    cli: argparse.Namespace
    instance_config: sambacc.config.InstanceConfig

    def __init__(self, opts, instance_config):
        self.cli = argparse.Namespace()
        self.instance_config = instance_config
        for k, v in opts.items():
            setattr(self.cli, k, v)
        self.require_validation = False

    @classmethod
    def defaults(cls, cfg_path, watch=False):
        with open(cfg_path, "w") as fh:
            fh.write(config1)

        config = [cfg_path]
        identity = "updateme"
        ctx = cls(
            {
                "watch": watch,
                "config": config,
                "identity": identity,
            },
            sambacc.config.read_config_files(config).get(identity),
        )
        return ctx

    @property
    def opener(self) -> sambacc.opener.Opener:
        return sambacc.opener.FileOpener()


class FakeWaiter:
    def __init__(self, attempts=None):
        self.count = 0
        self.on_count = {}
        self.attempts = attempts

    def acted(self):
        pass

    def wait(self):
        if self.attempts is not None and self.count >= self.attempts:
            raise KeyboardInterrupt()
        wf = self.on_count.get(self.count, None)
        if wf is not None:
            wf()
        self.count += 1


def _gen_fake_cmd(fake_path, chkpath, pnn="0"):
    with open(fake_path, "w") as fh:
        fh.write("#!/bin/sh\n")
        fh.write(f'echo "$@" >> {chkpath}\n')
        fh.write(f'[ "$1" = ctdb ] && echo {pnn}" " ; \n')
        fh.write("exit 0\n")
    os.chmod(fake_path, 0o755)


def test_update_config_changed(tmp_path, monkeypatch):
    cfg_path = str(tmp_path / "config")
    fake = tmp_path / "fake.sh"
    chkpath = tmp_path / ".executed"
    _gen_fake_cmd(fake, str(chkpath))
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", [str(fake)])

    ctx = FakeContext.defaults(cfg_path)
    with open(cfg_path, "w") as fh:
        fh.write(config2)
    monkeypatch.setattr(
        sambacc.paths,
        "ensure_share_dirs",
        functools.partial(
            sambacc.paths.ensure_share_dirs, root=str(tmp_path / "_root")
        ),
    )
    sambacc.commands.config.update_config(ctx)

    assert os.path.exists(chkpath)
    chk = open(chkpath).readlines()
    assert any(("net" in line) for line in chk)
    assert any(("smbcontrol" in line) for line in chk)


def test_update_config_changed_ctdb(tmp_path, monkeypatch):
    cfg_path = str(tmp_path / "config")
    fake = tmp_path / "fake.sh"
    chkpath = tmp_path / ".executed"
    _gen_fake_cmd(fake, str(chkpath))
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", [str(fake)])

    ctx = FakeContext.defaults(cfg_path)
    ctx.instance_config.iconfig["instance_features"] = ["ctdb"]
    assert ctx.instance_config.with_ctdb
    with open(cfg_path, "w") as fh:
        fh.write(config2)
    monkeypatch.setattr(
        sambacc.paths,
        "ensure_share_dirs",
        functools.partial(
            sambacc.paths.ensure_share_dirs, root=str(tmp_path / "_root")
        ),
    )
    sambacc.commands.config.update_config(ctx)

    assert os.path.exists(chkpath)
    chk = open(chkpath).readlines()
    assert any(("net" in line) for line in chk)
    assert any(("smbcontrol" in line) for line in chk)


def test_update_config_ctdb_notleader(tmp_path, monkeypatch):
    cfg_path = str(tmp_path / "config")
    fake = tmp_path / "fake.sh"
    chkpath = tmp_path / ".executed"
    _gen_fake_cmd(fake, str(chkpath), pnn="")
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", [str(fake)])

    ctx = FakeContext.defaults(cfg_path)
    ctx.instance_config.iconfig["instance_features"] = ["ctdb"]
    assert ctx.instance_config.with_ctdb
    with open(cfg_path, "w") as fh:
        fh.write(config2)
    monkeypatch.setattr(
        sambacc.paths,
        "ensure_share_dirs",
        functools.partial(
            sambacc.paths.ensure_share_dirs, root=str(tmp_path / "_root")
        ),
    )
    sambacc.commands.config.update_config(ctx)

    assert os.path.exists(chkpath)
    chk = open(chkpath).readlines()
    assert not any(("net" in line) for line in chk)
    assert not any(("smbcontrol" in line) for line in chk)


def test_update_config_watch_waiter_expires(tmp_path, monkeypatch):
    cfg_path = str(tmp_path / "config")
    fake = tmp_path / "fake.sh"
    chkpath = tmp_path / ".executed"
    _gen_fake_cmd(fake, str(chkpath))
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", [str(fake)])

    fake_waiter = FakeWaiter(attempts=5)

    def _fake_waiter(*args, **kwargs):
        return fake_waiter

    monkeypatch.setattr(sambacc.commands.config, "best_waiter", _fake_waiter)

    ctx = FakeContext.defaults(cfg_path, watch=True)
    monkeypatch.setattr(
        sambacc.paths,
        "ensure_share_dirs",
        functools.partial(
            sambacc.paths.ensure_share_dirs, root=str(tmp_path / "_root")
        ),
    )
    sambacc.commands.config.update_config(ctx)

    assert not os.path.exists(chkpath)
    assert fake_waiter.count == 5


def test_update_config_watch_waiter_trigger3(tmp_path, monkeypatch):
    cfg_path = str(tmp_path / "config")
    fake = tmp_path / "fake.sh"
    chkpath = tmp_path / ".executed"
    _gen_fake_cmd(fake, str(chkpath))
    monkeypatch.setattr(sambacc.samba_cmds, "_GLOBAL_PREFIX", [str(fake)])

    fake_waiter = FakeWaiter(attempts=5)

    def _fake_waiter(*args, **kwargs):
        return fake_waiter

    def _new_conf():
        with open(cfg_path, "w") as fh:
            fh.write(config2)

    monkeypatch.setattr(sambacc.commands.config, "best_waiter", _fake_waiter)
    fake_waiter.on_count[3] = _new_conf

    ctx = FakeContext.defaults(cfg_path, watch=True)
    monkeypatch.setattr(
        sambacc.paths,
        "ensure_share_dirs",
        functools.partial(
            sambacc.paths.ensure_share_dirs, root=str(tmp_path / "_root")
        ),
    )
    sambacc.commands.config.update_config(ctx)

    assert os.path.exists(chkpath)
    chk = open(chkpath).readlines()
    assert any(("net" in line) for line in chk)
    assert any(("smbcontrol" in line) for line in chk)
    assert fake_waiter.count == 5



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import errno
import os


def ensure_samba_dirs(root: str = "/") -> None:
    """Ensure that certain directories that samba servers expect will
    exist. This is useful when mapping iniitally empty dirs into
    the container.
    """
    smb_dir = os.path.join(root, "var/lib/samba")
    smb_private_dir = os.path.join(smb_dir, "private")
    smb_run_dir = os.path.join(root, "run/samba")
    wb_sockets_dir = os.path.join(smb_run_dir, "winbindd")

    _mkdir(smb_dir)
    _mkdir(smb_private_dir)

    _mkdir(smb_run_dir)
    _mkdir(wb_sockets_dir)
    os.chmod(wb_sockets_dir, 0o755)


def _mkdir(path: str) -> None:
    try:
        os.mkdir(path)
    except OSError as err:
        if getattr(err, "errno", 0) != errno.EEXIST:
            raise


def ensure_share_dirs(path: str, root: str = "/") -> None:
    """Ensure that the given path exists.
    The optional root argument allows "reparenting" the path
    into a virtual root dir.
    """
    while path.startswith("/"):
        path = path[1:]
    path = os.path.join(root, path)
    os.makedirs(path, exist_ok=True)



#
# sambacc: a samba container configuration tool
# Copyright (C) 2021  John Mulligan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
#

import argparse
import json
import logging
import os
import time
import typing

from sambacc import config
from sambacc import opener
from sambacc import rados_opener
from sambacc import samba_cmds
from sambacc import url_opener

from . import check  # noqa: F401
from . import config as config_cmds
from . import ctdb  # noqa: F401
from . import dns  # noqa: F401
from . import initialize  # noqa: F401
from . import join  # noqa: F401
from . import run  # noqa: F401
from . import skips
from . import users  # noqa: F401
from .cli import commands, Fail, Parser

DEFAULT_CONFIG = "/etc/samba/container/config.json"
DEFAULT_JOIN_MARKER = "/var/lib/samba/container-join-marker.json"

default_cfunc = config_cmds.print_config


def global_args(parser: Parser) -> None:
    parser.add_argument(
        "--config",
        action="append",
        help=(
            "Specify source configuration"
            " (can also be set in the environment by SAMBACC_CONFIG)."
        ),
    )
    parser.add_argument(
        "--identity",
        help=(
            "A string identifying the local identity"
            " (can also be set in the environment by SAMBA_CONTAINER_ID)."
        ),
    )
    parser.add_argument(
        "--etc-passwd-path",
        default="/etc/passwd",
        help="Specify a path for the passwd file.",
    )
    parser.add_argument(
        "--etc-group-path",
        default="/etc/group",
        help="Specify a path for the group file.",
    )
    parser.add_argument(
        "--username",
        default="Administrator",
        help="Specify a user name for domain access.",
    )
    parser.add_argument(
        "--password", default="", help="Specify a password for domain access."
    )
    parser.add_argument(
        "--debug-delay",
        type=int,
        help="Delay activity for a specified number of seconds.",
    )
    parser.add_argument(
        "--join-marker",
        default=DEFAULT_JOIN_MARKER,
        help="Path to a file used to indicate a join has been peformed.",
    )
    parser.add_argument(
        "--samba-debug-level",
        choices=[str(v) for v in range(0, 11)],
        help="Specify samba debug level for commands.",
    )
    parser.add_argument(
        "--samba-command-prefix",
        help="Wrap samba commands within a supplied command prefix",
    )
    parser.add_argument(
        "--skip-if",
        dest="skip_conditions",
        action="append",
        type=skips.parse,
        help=(
            "Skip execution based on a condition. Conditions include"
            " 'file:[!]<path>', 'env:<var>(==|!=)<value>', and 'always:'."
            " (Pass `?` for more details)"
        ),
    )
    parser.add_argument(
        "--skip-if-file",
        action="append",
        dest="skip_conditions",
        type=skips.SkipFile.parse,
        help="(DEPRECATED) Perform no action if the specified path exists.",
    )
    parser.add_argument(
        "--validate-config",
        choices=("auto", "required", "true", "false"),
        help="Perform schema based validation of configuration.",
    )
    parser.add_argument(
        "--ceph-id",
        type=_ceph_id,
        help=(
            "Specify a user/client ID to ceph libraries"
            "(can also be set in the environment by SAMBACC_CEPH_ID."
            " Ignored if Ceph RADOS libraries are not present or unused."
            " Pass `?` for more details)."
        ),
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug level logging of sambacc.",
    )


def _ceph_id(
    value: typing.Union[str, dict[str, typing.Any]]
) -> dict[str, typing.Any]:
    if not isinstance(value, str):
        return value
    if value == "?":
        # A hack to avoid putting tons of ceph specific info in the normal
        # help output. There's probably a better way to do this but it
        # gets the job done for now.
        raise argparse.ArgumentTypeError(
            "requested help:"
            " Specify names in the form"
            " --ceph-id=[key=value][,key=value][,...]."
            ' Valid keys include "name" to set the exact name and "rados_id"'
            ' to specify a name that lacks the "client." prefix (that will'
            "automatically get added)."
            " Alternatively, specify just the name to allow the system to"
            " guess if the name is prefixed already or not."
        )
    result: dict[str, typing.Any] = {}
    # complex mode
    if "=" in value:
        for part in value.split(","):
            if "=" not in part:
                raise argparse.ArgumentTypeError(
                    f"unexpected value for ceph-id: {value!r}"
                )
            key, val = part.split("=", 1)
            if key == "name":
                result["client_name"] = val
                result["full_name"] = True
            elif key == "rados_id":
                result["client_name"] = val
                result["full_name"] = False
            else:
                b = f"unexpected key {key!r} in value for ceph-id: {value!r}"
                raise argparse.ArgumentTypeError(b)
    else:
        # this shorthand is meant mainly for lazy humans (me) when running test
        # images manually. The key-value form above is meant for automation.
        result["client_name"] = value
        # assume that if the name starts with client. it's the full name and
        # avoid having the ceph library double up an create client.client.x.
        result["full_name"] = value.startswith("client.")
    return result


def from_env(
    ns: typing.Any,
    var: str,
    ename: str,
    default: typing.Any = None,
    convert_env: typing.Optional[typing.Callable] = None,
    convert_value: typing.Optional[typing.Callable] = str,
) -> None:
    value = getattr(ns, var, None)
    if not value:
        value = os.environ.get(ename, "")
        if convert_env is not None:
            value = convert_env(value)
    if convert_value is not None:
        value = convert_value(value)
    if value:
        setattr(ns, var, value)


def split_entries(value):
    out = []
    if not isinstance(value, str):
        raise ValueError(value)
    if not value:
        return out
    # in order to cleanly allow passing uris as config "paths" we can't
    # simply split on colons. Avoid coming up with a hokey custom scheme
    # and enter "JSON-mode" if the env var starts and ends with brackets
    # hinting it contains a JSON list.
    v = value.rstrip(None)  # permit trailing whitespace (trailing only!)
    if v[0] == "[" and v[-1] == "]":
        for item in json.loads(v):
            if not isinstance(item, str):
                raise ValueError("Variable JSON must be a list of strings")
            out.append(item)
    else:
        # backwards compatibilty mode with `PATH` like syntax
        for part in value.split(":"):
            out.append(part)
    return out


def env_to_cli(cli: typing.Any) -> None:
    from_env(
        cli,
        "config",
        "SAMBACC_CONFIG",
        convert_env=split_entries,
        convert_value=None,
        default=DEFAULT_CONFIG,
    )
    from_env(
        cli,
        "join_files",
        "SAMBACC_JOIN_FILES",
        convert_env=split_entries,
        convert_value=None,
    )
    from_env(cli, "identity", "SAMBA_CONTAINER_ID")
    from_env(cli, "username", "JOIN_USERNAME")
    from_env(cli, "password", "INSECURE_JOIN_PASSWORD")
    from_env(cli, "samba_debug_level", "SAMBA_DEBUG_LEVEL")
    from_env(cli, "validate_config", "SAMBACC_VALIDATE_CONFIG")
    from_env(cli, "ceph_id", "SAMBACC_CEPH_ID", convert_value=_ceph_id)


class CommandContext:
    """CLI Context for standard samba-container commands."""

    def __init__(self, cli_args: argparse.Namespace):
        self._cli = cli_args
        self._iconfig: typing.Optional[config.InstanceConfig] = None
        self.expects_ctdb = False
        self._opener: typing.Optional[opener.Opener] = None

    @property
    def cli(self) -> argparse.Namespace:
        return self._cli

    @property
    def instance_config(self) -> config.InstanceConfig:
        if self._iconfig is None:
            cfgs = self.cli.config or []
            self._iconfig = config.read_config_files(
                cfgs,
                require_validation=self.require_validation,
                opener=self.opener,
            ).get(self.cli.identity)
        return self._iconfig

    @property
    def require_validation(self) -> typing.Optional[bool]:
        if self.cli.validate_config in ("required", "true"):
            return True
        if self.cli.validate_config == "false":
            return False
        return None

    @property
    def opener(self) -> opener.Opener:
        if self._opener is None:
            self._opener = opener.FallbackOpener([url_opener.URLOpener()])
        return self._opener


def pre_action(cli: typing.Any) -> None:
    """Handle debugging/diagnostic related options before the target
    action of the command is performed.
    """
    if cli.debug_delay:
        time.sleep(int(cli.debug_delay))
    if cli.samba_debug_level:
        samba_cmds.set_global_debug(cli.samba_debug_level)
    if cli.samba_command_prefix:
        samba_cmds.set_global_prefix([cli.samba_command_prefix])

    # should there be an option to force {en,dis}able rados?
    # Right now we just always try to enable rados when possible.
    rados_opener.enable_rados(
        url_opener.URLOpener,
        client_name=cli.ceph_id.get("client_name", ""),
        full_name=cli.ceph_id.get("full_name", False),
    )


def enable_logging(cli: typing.Any) -> None:
    level = logging.DEBUG if cli.debug else logging.INFO
    logger = logging.getLogger()
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("{asctime}: {levelname}: {message}", style="{")
    )
    handler.setLevel(level)
    logger.addHandler(handler)


def main(args: typing.Optional[typing.Sequence[str]] = None) -> None:
    cli = commands.assemble(arg_func=global_args).parse_args(args)
    env_to_cli(cli)
    enable_logging(cli)
    if not cli.identity:
        raise Fail("missing container identity")

    pre_action(cli)
    ctx = CommandContext(cli)
    skip = skips.test(ctx)
    if skip:
        print(f"Command Skipped: {skip}")
        return
    cfunc = getattr(cli, "cfunc", default_cfunc)
    cfunc(ctx)
    return


if __name__ == "__main__":
    main()
